@@ -36,6 +36,9 @@
 #define DEF_FILE_SIZE		(1024 * 1024 * 1024UL)
 #define DEF_ZONE_SIZE		(0)
 #define DEF_ZONE_SKIP		(0)
+#define DEF_RWMIX_CYCLE		(500)
+#define DEF_RWMIX_READ		(50)
+#define DEF_NICE		(0)
 
 static char fio_version_string[] = "fio 1.3";
 
@@ -50,6 +53,7 @@ int rate_quit = 0;
 int write_lat_log = 0;
 int write_bw_log = 0;
 int exitall_on_terminate = 0;
+unsigned long long mlock_size = 0;
 
 static int setup_rate(struct thread_data *td)
 {
@@ -81,11 +85,11 @@ static void setup_log(struct io_log **log)
 
 void finish_log(struct thread_data *td, struct io_log *log, const char *name)
 {
-	char file_name[128];
+	char file_name[256];
 	FILE *f;
 	unsigned int i;
 
-	sprintf(file_name, "client%d_%s.log", td->thread_number, name);
+	snprintf(file_name, 200, "client%d_%s.log", td->thread_number, name);
 	f = fopen(file_name, "w");
 	if (!f) {
 		perror("fopen log");
@@ -110,7 +114,10 @@ static struct thread_data *get_new_job(int global, struct thread_data *parent)
 		return NULL;
 
 	td = &threads[thread_number++];
-	memset(td, 0, sizeof(*td));
+	if (parent)
+		*td = *parent;
+	else
+		memset(td, 0, sizeof(*td));
 
 	td->fd = -1;
 	td->thread_number = thread_number;
@@ -161,10 +168,10 @@ static void put_job(struct thread_data *td)
 	thread_number--;
 }
 
-static int add_job(struct thread_data *td, const char *jobname, int prioclass,
-		   int prio)
+static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 {
-	char *ddir_str[] = { "read", "write", "randread", "randwrite" };
+	char *ddir_str[] = { "read", "write", "randread", "randwrite",
+			     "rw", NULL, "randrw" };
 	struct stat sb;
 	int numjobs, ddir;
 
@@ -180,9 +187,6 @@ static int add_job(struct thread_data *td, const char *jobname, int prioclass,
 		return 1;
 	}
 #endif
-#ifdef FIO_HAVE_IOPRIO
-	td->ioprio = (prioclass << IOPRIO_CLASS_SHIFT) | prio;
-#endif
 
 	/*
 	 * the def_thread is just for options, it's not a real job
@@ -213,11 +217,11 @@ static int add_job(struct thread_data *td, const char *jobname, int prioclass,
 
 	if (td->filetype == FIO_TYPE_FILE) {
 		if (td->directory[0] != '\0')
-			sprintf(td->file_name, "%s/%s.%d", td->directory, jobname, td->thread_number);
+			sprintf(td->file_name, "%s/%s.%d", td->directory, jobname, td->jobnum);
 		else
-			sprintf(td->file_name, "%s.%d", jobname, td->thread_number);
+			sprintf(td->file_name, "%s.%d", jobname, td->jobnum);
 	} else
-		strcpy(td->file_name, jobname);
+		strncpy(td->file_name, jobname, sizeof(td->file_name) - 1);
 
 	sem_init(&td->mutex, 0, 0);
 
@@ -229,7 +233,7 @@ static int add_job(struct thread_data *td, const char *jobname, int prioclass,
 		td->min_bs = td->bs;
 	if (td->max_bs == -1U)
 		td->max_bs = td->bs;
-	if (td_read(td))
+	if (td_read(td) && !td_rw(td))
 		td->verify = 0;
 
 	if (td->stonewall && td->thread_number > 1)
@@ -247,8 +251,12 @@ static int add_job(struct thread_data *td, const char *jobname, int prioclass,
 	if (write_bw_log)
 		setup_log(&td->bw_log);
 
-	ddir = td->ddir + (!td->sequential << 1);
-	printf("Client%d (g=%d): rw=%s, prio=%d/%d, odir=%d, bs=%d-%d, rate=%d, ioengine=%s, iodepth=%d\n", td->thread_number, td->groupid, ddir_str[ddir], prioclass, prio, td->odirect, td->min_bs, td->max_bs, td->rate, td->io_engine_name, td->iodepth);
+	ddir = td->ddir + (!td->sequential << 1) + (td->iomix << 2);
+
+	if (!job_add_num)
+		printf("Client%d (g=%d): rw=%s, odir=%d, bs=%d-%d, rate=%d, ioengine=%s, iodepth=%d\n", td->thread_number, td->groupid, ddir_str[ddir], td->odirect, td->min_bs, td->max_bs, td->rate, td->io_engine_name, td->iodepth);
+	else if (job_add_num == 1)
+		printf("...\n");
 
 	/*
 	 * recurse add identical jobs, clear numjobs and stonewall options
@@ -263,8 +271,10 @@ static int add_job(struct thread_data *td, const char *jobname, int prioclass,
 
 		td_new->numjobs = 1;
 		td_new->stonewall = 0;
+		td_new->jobnum = numjobs;
+		job_add_num = numjobs - 1;
 
-		if (add_job(td_new, jobname, prioclass, prio))
+		if (add_job(td_new, jobname, job_add_num))
 			goto err;
 	}
 	return 0;
@@ -275,16 +285,16 @@ err:
 
 int init_random_state(struct thread_data *td)
 {
-	unsigned long seed;
+	unsigned long seeds[4];
 	int fd, num_maps, blocks;
 
-	fd = open("/dev/random", O_RDONLY);
+	fd = open("/dev/urandom", O_RDONLY);
 	if (fd == -1) {
 		td_verror(td, errno);
 		return 1;
 	}
 
-	if (read(fd, &seed, sizeof(seed)) < (int) sizeof(seed)) {
+	if (read(fd, seeds, sizeof(seeds)) < (int) sizeof(seeds)) {
 		td_verror(td, EIO);
 		close(fd);
 		return 1;
@@ -292,14 +302,15 @@ int init_random_state(struct thread_data *td)
 
 	close(fd);
 
-	srand48_r(seed, &td->bsrange_state);
-	srand48_r(seed, &td->verify_state);
+	srand48_r(seeds[0], &td->bsrange_state);
+	srand48_r(seeds[1], &td->verify_state);
+	srand48_r(seeds[2], &td->rwmix_state);
 
 	if (td->sequential)
 		return 0;
 
 	if (repeatable)
-		seed = DEF_RANDSEED;
+		seeds[3] = DEF_RANDSEED;
 
 	blocks = (td->io_size + td->min_bs - 1) / td->min_bs;
 	num_maps = blocks / BLOCKS_PER_MAP;
@@ -307,7 +318,7 @@ int init_random_state(struct thread_data *td)
 	td->num_maps = num_maps;
 	memset(td->file_map, 0, num_maps * sizeof(long));
 
-	srand48_r(seed, &td->random_state);
+	srand48_r(seeds[3], &td->random_state);
 	return 0;
 }
 
@@ -361,7 +372,7 @@ static int str_cnv(char *p, unsigned long long *val)
 	if (*val == ULONG_MAX && errno == ERANGE)
 		return 1;
 
-	*val *= get_mult(str[len - 2]);
+	*val *= get_mult(str[len - 1]);
 	return 0;
 }
 
@@ -377,24 +388,30 @@ static void strip_blank_front(char **p)
 {
 	char *s = *p;
 
-	while (isblank(*s))
+	while (isspace(*s))
 		s++;
 }
 
 static void strip_blank_end(char *p)
 {
-	while (isblank(*p)) {
-		*p = '\0';
-		p--;
-	}
+	char *s = p + strlen(p) - 1;
+
+	while (isspace(*s) || iscntrl(*s))
+		s--;
+
+	*(s + 1) = '\0';
 }
 
 typedef int (str_cb_fn)(struct thread_data *, char *);
 
 static int check_str(char *p, char *name, str_cb_fn *cb, struct thread_data *td)
 {
-	char *s = strstr(p, name);
+	char *s;
 
+	if (strncmp(p, name, strlen(name)))
+		return 1;
+
+	s = strstr(p, name);
 	if (!s)
 		return 1;
 
@@ -409,8 +426,12 @@ static int check_str(char *p, char *name, str_cb_fn *cb, struct thread_data *td)
 
 static int check_strstore(char *p, char *name, char *dest)
 {
-	char *s = strstr(p, name);
+	char *s;
 
+	if (strncmp(p, name, strlen(name)))
+		return 1;
+
+	s = strstr(p, name);
 	if (!s)
 		return 1;
 
@@ -422,9 +443,6 @@ static int check_strstore(char *p, char *name, char *dest)
 	strip_blank_front(&s);
 
 	strcpy(dest, s);
-
-	s = dest + strlen(dest) - 1;
-	strip_blank_end(s);
 	return 0;
 }
 
@@ -448,6 +466,9 @@ static int check_range(char *p, char *name, unsigned long *s, unsigned long *e)
 	char option[128];
 	char *str, *p1, *p2;
 
+	if (strncmp(p, name, strlen(name)))
+		return 1;
+
 	strcpy(option, p);
 	p = option;
 
@@ -486,6 +507,9 @@ static int check_int(char *p, char *name, unsigned int *val)
 {
 	char *str;
 
+	if (strncmp(p, name, strlen(name)))
+		return 1;
+
 	str = strstr(p, name);
 	if (!str)
 		return 1;
@@ -539,6 +563,16 @@ static int str_rw_cb(struct thread_data *td, char *mem)
 		td->ddir = DDIR_WRITE;
 		td->sequential = 0;
 		return 0;
+	} else if (!strncmp(mem, "rw", 2)) {
+		td->ddir = 0;
+		td->iomix = 1;
+		td->sequential = 1;
+		return 0;
+	} else if (!strncmp(mem, "randrw", 6)) {
+		td->ddir = 0;
+		td->iomix = 1;
+		td->sequential = 0;
+		return 0;
 	}
 
 	fprintf(stderr, "bad data direction: %s\n", mem);
@@ -602,16 +636,25 @@ static int str_ioengine_cb(struct thread_data *td, char *str)
 		strcpy(td->io_engine_name, "sgio");
 		td->io_engine = FIO_SGIO;
 		return 0;
+	} else if (!strncmp(str, "splice", 6)) {
+		strcpy(td->io_engine_name, "splice");
+		td->io_engine = FIO_SPLICEIO;
+		return 0;
 	}
 
 	fprintf(stderr, "bad ioengine type: %s\n", str);
 	return 1;
 }
 
+static int str_iolog_cb(struct thread_data *td, char *file)
+{
+	strncpy(td->iolog_file, file, sizeof(td->iolog_file) - 1);
+	return 0;
+}
 
 int parse_jobs_ini(char *file)
 {
-	unsigned int prioclass, prio, cpu, global;
+	unsigned int prioclass, prio, cpu, global, il;
 	unsigned long long ull;
 	unsigned long ul1, ul2;
 	struct thread_data *td;
@@ -622,7 +665,7 @@ int parse_jobs_ini(char *file)
 
 	f = fopen(file, "r");
 	if (!f) {
-		perror("fopen");
+		perror("fopen job file");
 		return 1;
 	}
 
@@ -643,20 +686,21 @@ int parse_jobs_ini(char *file)
 		if (!td)
 			return 1;
 
-		prioclass = 2;
-		prio = 4;
-
 		fgetpos(f, &off);
 		while ((p = fgets(string, 4096, f)) != NULL) {
 			if (is_empty_or_comment(p))
 				continue;
 			if (strstr(p, "["))
 				break;
+			strip_blank_front(&p);
+			strip_blank_end(p);
+
 			if (!check_int(p, "prio", &prio)) {
 #ifndef FIO_HAVE_IOPRIO
 				fprintf(stderr, "io priorities not available\n");
 				return 1;
 #endif
+				td->ioprio |= prio;
 				fgetpos(f, &off);
 				continue;
 			}
@@ -665,6 +709,7 @@ int parse_jobs_ini(char *file)
 				fprintf(stderr, "io priorities not available\n");
 				return 1;
 #endif
+				td->ioprio |= prioclass << IOPRIO_CLASS_SHIFT;
 				fgetpos(f, &off);
 				continue;
 			}
@@ -699,6 +744,7 @@ int parse_jobs_ini(char *file)
 			}
 			if (!check_int(p, "fsync", &td->fsync_blocks)) {
 				fgetpos(f, &off);
+				td->end_fsync = 1;
 				continue;
 			}
 			if (!check_int(p, "startdelay", &td->start_delay)) {
@@ -733,6 +779,10 @@ int parse_jobs_ini(char *file)
 				fgetpos(f, &off);
 				continue;
 			}
+			if (!check_int(p, "end_fsync", &td->end_fsync)) {
+				fgetpos(f, &off);
+				continue;
+			}
 			if (!check_int(p, "loops", &td->loops)) {
 				fgetpos(f, &off);
 				continue;
@@ -745,6 +795,28 @@ int parse_jobs_ini(char *file)
 				fgetpos(f, &off);
 				continue;
 			}
+			if (!check_int(p, "rwmixcycle", &td->rwmixcycle)) {
+				fgetpos(f, &off);
+				continue;
+			}
+			if (!check_int(p, "rwmixread", &il)) {
+				if (il > 100)
+					il = 100;
+				td->rwmixread = il;
+				fgetpos(f, &off);
+				continue;
+			}
+			if (!check_int(p, "rwmixwrite", &il)) {
+				if (il > 100)
+					il = 100;
+				td->rwmixread = 100 - il;
+				fgetpos(f, &off);
+				continue;
+			}
+			if (!check_int(p, "nice", &td->nice)) {
+				fgetpos(f, &off);
+				continue;
+			}
 			if (!check_range(p, "bsrange", &ul1, &ul2)) {
 				if (ul1 > ul2) {
 					td->max_bs = ul1;
@@ -777,6 +849,10 @@ int parse_jobs_ini(char *file)
 				fgetpos(f, &off);
 				continue;
 			}
+			if (!check_strcnv(p, "lockmem", &mlock_size)) {
+				fgetpos(f, &off);
+				continue;
+			}
 			if (!check_strstore(p, "directory", td->directory)) {
 				fgetpos(f, &off);
 				continue;
@@ -817,12 +893,27 @@ int parse_jobs_ini(char *file)
 				fgetpos(f, &off);
 				continue;
 			}
+			if (!check_str(p, "iolog", str_iolog_cb, td)) {
+				printf("got read iolog\n");
+				td->read_iolog = 1;
+				td->write_iolog = 0;
+				fgetpos(f, &off);
+				continue;
+			}
+			if (!td->read_iolog &&
+			    !check_str(p, "write_iolog", str_iolog_cb, td)) {
+				printf("got write iolog\n");
+				td->write_iolog = 1;
+				fgetpos(f, &off);
+				continue;
+			}
 
 			printf("Client%d: bad option %s\n",td->thread_number,p);
+			return 1;
 		}
 		fsetpos(f, &off);
 
-		if (add_job(td, name, prioclass, prio))
+		if (add_job(td, name, 0))
 			return 1;
 	}
 
@@ -845,6 +936,7 @@ static int fill_def_thread(void)
 	 * fill globals
 	 */
 	def_thread.ddir = DDIR_READ;
+	def_thread.iomix = 0;
 	def_thread.bs = DEF_BS;
 	def_thread.min_bs = -1;
 	def_thread.max_bs = -1;
@@ -867,6 +959,9 @@ static int fill_def_thread(void)
 	def_thread.stonewall = DEF_STONEWALL;
 	def_thread.numjobs = DEF_NUMJOBS;
 	def_thread.use_thread = DEF_USE_THREAD;
+	def_thread.rwmixcycle = DEF_RWMIX_CYCLE;
+	def_thread.rwmixread = DEF_RWMIX_READ;
+	def_thread.nice = DEF_NICE;
 #ifdef FIO_HAVE_DISK_UTIL
 	def_thread.do_disk_util = 1;
 #endif
@@ -874,11 +969,25 @@ static int fill_def_thread(void)
 	return 0;
 }
 
+static void usage(char *name)
+{
+	printf("%s\n", fio_version_string);
+	printf("\t-s IO is sequential\n");
+	printf("\t-b Block size in KiB for each IO\n");
+	printf("\t-t Runtime in seconds\n");
+	printf("\t-R Exit all threads on failure to meet rate goal\n");
+	printf("\t-o Use O_DIRECT\n");
+	printf("\t-l Generate per-job latency logs\n");
+	printf("\t-w Generate per-job bandwidth logs\n");
+	printf("\t-f Job file (Required)\n");
+	printf("\t-v Print version info and exit\n");
+}
+
 static void parse_cmd_line(int argc, char *argv[])
 {
 	int c;
 
-	while ((c = getopt(argc, argv, "s:b:t:r:R:o:f:lwv")) != EOF) {
+	while ((c = getopt(argc, argv, "s:b:t:r:R:o:f:lwvh")) != EOF) {
 		switch (c) {
 			case 's':
 				def_thread.sequential = !!atoi(optarg);
@@ -912,11 +1021,17 @@ static void parse_cmd_line(int argc, char *argv[])
 			case 'w':
 				write_bw_log = 1;
 				break;
+			case 'h':
+				usage(argv[0]);
+				exit(0);
 			case 'v':
 				printf("%s\n", fio_version_string);
 				exit(0);
 		}
 	}
+
+	if (!ini_file && argc > 1 && argv[argc - 1][0] != '-')
+		ini_file = strdup(argv[argc - 1]);
 }
 
 static void free_shm(void)
@@ -974,11 +1089,14 @@ int parse_options(int argc, char *argv[])
 
 	if (!ini_file) {
 		printf("Need job file\n");
+		usage(argv[0]);
 		return 1;
 	}
 
-	if (parse_jobs_ini(ini_file))
+	if (parse_jobs_ini(ini_file)) {
+		usage(argv[0]);
 		return 1;
+	}
 
 	return 0;
 }
