@@ -57,23 +57,27 @@ static void update_io_ticks(void);
 static void disk_util_timer_arm(void);
 static void print_thread_status(void);
 
+extern unsigned long long mlock_size;
+
 /*
  * thread life cycle
  */
 enum {
 	TD_NOT_CREATED = 0,
 	TD_CREATED,
+	TD_INITIALIZED,
 	TD_RUNNING,
 	TD_VERIFYING,
 	TD_EXITED,
 	TD_REAPED,
 };
 
-#define should_fsync(td)	(td_write(td) && (!(td)->odirect || (td)->override_sync))
+#define should_fsync(td)	((td_write(td) || td_rw(td)) && (!(td)->odirect || (td)->override_sync))
 
 static sem_t startup_sem;
 
 #define TERMINATE_ALL		(-1)
+#define JOB_START_TIMEOUT	(5 * 1000)
 
 static void terminate_threads(int group_id)
 {
@@ -193,7 +197,7 @@ static int get_next_free_block(struct thread_data *td, unsigned long long *b)
 
 static void mark_random_map(struct thread_data *td, struct io_u *io_u)
 {
-	unsigned long block = io_u->offset / td->min_bs;
+	unsigned long long block = io_u->offset / (unsigned long long) td->min_bs;
 	unsigned int blocks = 0;
 
 	while (blocks < (io_u->buflen / td->min_bs)) {
@@ -284,12 +288,12 @@ static int get_next_offset(struct thread_data *td, unsigned long long *offset)
 	long r;
 
 	if (!td->sequential) {
-		unsigned long max_blocks = td->io_size / td->min_bs;
+		unsigned long long max_blocks = td->io_size / td->min_bs;
 		int loops = 50;
 
 		do {
 			lrand48_r(&td->random_state, &r);
-			b = ((max_blocks - 1) * r / (RAND_MAX+1.0));
+			b = ((max_blocks - 1) * r / (unsigned long long) (RAND_MAX+1.0));
 			rb = b + (td->file_offset / td->min_bs);
 			loops--;
 		} while (!random_map_free(td, rb) && loops);
@@ -541,6 +545,37 @@ static void fill_md5(struct verify_header *hdr, void *p, unsigned int len)
 	memcpy(hdr->md5_digest, md5_ctx.hash, sizeof(md5_ctx.hash));
 }
 
+static int get_rw_ddir(struct thread_data *td)
+{
+	if (td_rw(td)) {
+		struct timeval now;
+		unsigned long elapsed;
+
+		gettimeofday(&now, NULL);
+	 	elapsed = mtime_since_now(&td->rwmix_switch);
+
+		/*
+		 * Check if it's time to seed a new data direction.
+		 */
+		if (elapsed >= td->rwmixcycle) {
+			unsigned long v;
+			long r;
+
+			lrand48_r(&td->random_state, &r);
+			v = 100UL * r / (unsigned long) (RAND_MAX + 1.0);
+			if (v < td->rwmixread)
+				td->rwmix_ddir = DDIR_READ;
+			else
+				td->rwmix_ddir = DDIR_WRITE;
+			memcpy(&td->rwmix_switch, &now, sizeof(now));
+		}
+		return td->rwmix_ddir;
+	} else if (td_read(td))
+		return DDIR_READ;
+	else
+		return DDIR_WRITE;
+}
+
 /*
  * fill body of io_u->buf with random data and add a header with the
  * (eg) sha1sum of that data.
@@ -566,13 +601,8 @@ static void populate_io_u(struct thread_data *td, struct io_u *io_u)
 	memcpy(io_u->buf, &hdr, sizeof(hdr));
 }
 
-static int td_io_prep(struct thread_data *td, struct io_u *io_u, int read)
+static int td_io_prep(struct thread_data *td, struct io_u *io_u)
 {
-	if (read)
-		io_u->ddir = DDIR_READ;
-	else
-		io_u->ddir = DDIR_WRITE;
-
 	if (td->io_prep && td->io_prep(td, io_u))
 		return 1;
 
@@ -586,6 +616,58 @@ void put_io_u(struct thread_data *td, struct io_u *io_u)
 	td->cur_depth--;
 }
 
+static void write_iolog_put(struct thread_data *td, struct io_u *io_u)
+{
+	fprintf(td->iolog_f, "%d,%llu,%u\n", io_u->ddir, io_u->offset, io_u->buflen);
+}
+
+static int read_iolog_get(struct thread_data *td, struct io_u *io_u)
+{
+	struct io_piece *ipo;
+
+	if (!list_empty(&td->io_log_list)) {
+		ipo = list_entry(td->io_log_list.next, struct io_piece, list);
+		list_del(&ipo->list);
+		io_u->offset = ipo->offset;
+		io_u->buflen = ipo->len;
+		io_u->ddir = ipo->ddir;
+		free(ipo);
+		return 0;
+	}
+
+	return 1;
+}
+
+static int fill_io_u(struct thread_data *td, struct io_u *io_u)
+{
+	/*
+	 * If using an iolog, grab next piece if any available.
+	 */
+	if (td->read_iolog)
+		return read_iolog_get(td, io_u);
+
+	/*
+	 * No log, let the seq/rand engine retrieve the next position.
+	 */
+	if (!get_next_offset(td, &io_u->offset)) {
+		io_u->buflen = get_next_buflen(td);
+
+		if (io_u->buflen) {
+			io_u->ddir = get_rw_ddir(td);
+
+			/*
+			 * If using a write iolog, store this entry.
+			 */
+			if (td->write_iolog)
+				write_iolog_put(td, io_u);
+
+			return 0;
+		}
+	}
+
+	return 1;
+}
+
 #define queue_full(td)	(list_empty(&(td)->io_u_freelist))
 
 struct io_u *__get_io_u(struct thread_data *td)
@@ -617,13 +699,7 @@ static struct io_u *get_io_u(struct thread_data *td)
 		td->last_pos += td->zone_skip;
 	}
 
-	if (get_next_offset(td, &io_u->offset)) {
-		put_io_u(td, io_u);
-		return NULL;
-	}
-
-	io_u->buflen = get_next_buflen(td);
-	if (!io_u->buflen) {
+	if (fill_io_u(td, io_u)) {
 		put_io_u(td, io_u);
 		return NULL;
 	}
@@ -636,7 +712,7 @@ static struct io_u *get_io_u(struct thread_data *td)
 		return NULL;
 	}
 
-	if (!td->sequential)
+	if (!td->read_iolog && !td->sequential)
 		mark_random_map(td, io_u);
 
 	td->last_pos += io_u->buflen;
@@ -644,7 +720,7 @@ static struct io_u *get_io_u(struct thread_data *td)
 	if (td->verify != VERIFY_NONE)
 		populate_io_u(td, io_u);
 
-	if (td_io_prep(td, io_u, td_read(td))) {
+	if (td_io_prep(td, io_u)) {
 		put_io_u(td, io_u);
 		return NULL;
 	}
@@ -659,8 +735,7 @@ static inline void td_set_runstate(struct thread_data *td, int runstate)
 	td->runstate = runstate;
 }
 
-static int get_next_verify(struct thread_data *td,
-			   unsigned long long *offset, unsigned int *len)
+static int get_next_verify(struct thread_data *td, struct io_u *io_u)
 {
 	struct io_piece *ipo;
 
@@ -670,8 +745,9 @@ static int get_next_verify(struct thread_data *td,
 	ipo = list_entry(td->io_hist_list.next, struct io_piece, list);
 	list_del(&ipo->list);
 
-	*offset = ipo->offset;
-	*len = ipo->len;
+	io_u->offset = ipo->offset;
+	io_u->buflen = ipo->len;
+	io_u->ddir = DDIR_READ;
 	free(ipo);
 	return 0;
 }
@@ -724,6 +800,94 @@ static void log_io_piece(struct thread_data *td, struct io_u *io_u)
 	list_add(&ipo->list, entry);
 }
 
+static void write_iolog_close(struct thread_data *td)
+{
+	fflush(td->iolog_f);
+	fclose(td->iolog_f);
+	free(td->iolog_buf);
+}
+
+static int init_iolog(struct thread_data *td)
+{
+	unsigned long long offset;
+	unsigned int bytes;
+	char *str, *p;
+	FILE *f;
+	int rw, i, reads, writes;
+
+	if (!td->read_iolog && !td->write_iolog)
+		return 0;
+
+	if (td->read_iolog)
+		f = fopen(td->iolog_file, "r");
+	else
+		f = fopen(td->iolog_file, "w");
+
+	if (!f) {
+		perror("fopen iolog");
+		printf("file %s, %d/%d\n", td->iolog_file, td->read_iolog, td->write_iolog);
+		return 1;
+	}
+
+	/*
+	 * That's it for writing, setup a log buffer and we're done.
+	  */
+	if (td->write_iolog) {
+		td->iolog_f = f;
+		td->iolog_buf = malloc(8192);
+		setvbuf(f, td->iolog_buf, _IOFBF, 8192);
+		return 0;
+	}
+
+	/*
+	 * Read in the read iolog and store it, reuse the infrastructure
+	 * for doing verifications.
+	 */
+	str = malloc(4096);
+	reads = writes = i = 0;
+	while ((p = fgets(str, 4096, f)) != NULL) {
+		struct io_piece *ipo;
+
+		if (sscanf(p, "%d,%llu,%u", &rw, &offset, &bytes) != 3) {
+			fprintf(stderr, "bad iolog: %s\n", p);
+			continue;
+		}
+		if (rw == DDIR_READ)
+			reads++;
+		else if (rw == DDIR_WRITE)
+			writes++;
+		else {
+			fprintf(stderr, "bad ddir: %d\n", rw);
+			continue;
+		}
+
+		ipo = malloc(sizeof(*ipo));
+		INIT_LIST_HEAD(&ipo->list);
+		ipo->offset = offset;
+		ipo->len = bytes;
+		if (bytes > td->max_bs)
+			td->max_bs = bytes;
+		ipo->ddir = rw;
+		list_add_tail(&ipo->list, &td->io_log_list);
+		i++;
+	}
+
+	free(str);
+	fclose(f);
+
+	if (!i)
+		return 1;
+
+	if (reads && !writes)
+		td->ddir = DDIR_READ;
+	else if (!reads && writes)
+		td->ddir = DDIR_READ;
+	else
+		td->iomix = 1;
+
+	return 0;
+}
+
 static int sync_td(struct thread_data *td)
 {
 	if (td->io_sync)
@@ -757,7 +921,7 @@ static void io_completed(struct thread_data *td, struct io_u *io_u,
 
 	if (!io_u->error) {
 		unsigned int bytes = io_u->buflen - io_u->resid;
-		int idx = io_u->ddir;
+		const int idx = io_u->ddir;
 
 		td->io_blocks[idx]++;
 		td->io_bytes[idx] += bytes;
@@ -766,10 +930,10 @@ static void io_completed(struct thread_data *td, struct io_u *io_u,
 
 		msec = mtime_since(&io_u->issue_time, &e);
 
-		add_clat_sample(td, io_u->ddir, msec);
-		add_bw_sample(td, io_u->ddir);
+		add_clat_sample(td, idx, msec);
+		add_bw_sample(td, idx);
 
-		if (td_write(td) && io_u->ddir == DDIR_WRITE)
+		if ((td_rw(td) || td_write(td)) && idx == DDIR_WRITE)
 			log_io_piece(td, io_u);
 
 		icd->bytes_done[idx] += bytes;
@@ -867,12 +1031,12 @@ static void do_verify(struct thread_data *td)
 		if (!io_u)
 			break;
 
-		if (get_next_verify(td, &io_u->offset, &io_u->buflen)) {
+		if (get_next_verify(td, io_u)) {
 			put_io_u(td, io_u);
 			break;
 		}
 
-		if (td_io_prep(td, io_u, 1)) {
+		if (td_io_prep(td, io_u)) {
 			put_io_u(td, io_u);
 			break;
 		}
@@ -1007,7 +1171,7 @@ static void do_io(struct thread_data *td)
 	if (td->cur_depth)
 		cleanup_pending_aio(td);
 
-	if (should_fsync(td))
+	if (should_fsync(td) && td->end_fsync)
 		sync_td(td);
 }
 
@@ -1029,6 +1193,8 @@ static int init_io(struct thread_data *td)
 		return fio_posixaio_init(td);
 	else if (td->io_engine == FIO_SGIO)
 		return fio_sgio_init(td);
+	else if (td->io_engine == FIO_SPLICEIO)
+		return fio_spliceio_init(td);
 	else {
 		fprintf(stderr, "bad io_engine %d\n", td->io_engine);
 		return 1;
@@ -1102,10 +1268,6 @@ static int init_io_u(struct thread_data *td)
 		}
 	}
 
-	INIT_LIST_HEAD(&td->io_u_freelist);
-	INIT_LIST_HEAD(&td->io_u_busylist);
-	INIT_LIST_HEAD(&td->io_hist_list);
-
 	p = ALIGN(td->orig_buffer);
 	for (i = 0; i < max_units; i++) {
 		io_u = malloc(sizeof(*io_u));
@@ -1273,14 +1435,15 @@ static int setup_file_mmap(struct thread_data *td)
 {
 	int flags;
 
-	if (td_read(td))
-		flags = PROT_READ;
-	else {
+	if (td_rw(td))
+		flags = PROT_READ | PROT_WRITE;
+	else if (td_write(td)) {
 		flags = PROT_WRITE;
 
 		if (td->verify != VERIFY_NONE)
 			flags |= PROT_READ;
-	}
+	} else
+		flags = PROT_READ;
 
 	td->mmap = mmap(NULL, td->file_size, flags, MAP_SHARED, td->fd, td->file_offset);
 	if (td->mmap == MAP_FAILED) {
@@ -1361,14 +1524,7 @@ static int setup_file(struct thread_data *td)
 	if (td->odirect)
 		flags |= O_DIRECT;
 
-	if (td_read(td)) {
-		if (td->filetype == FIO_TYPE_CHAR)
-			flags |= O_RDWR;
-		else
-			flags |= O_RDONLY;
-
-		td->fd = open(td->file_name, flags);
-	} else {
+	if (td_write(td) || td_rw(td)) {
 		if (td->filetype == FIO_TYPE_FILE) {
 			if (!td->overwrite)
 				flags |= O_TRUNC;
@@ -1381,6 +1537,13 @@ static int setup_file(struct thread_data *td)
 		flags |= O_RDWR;
 
 		td->fd = open(td->file_name, flags, 0600);
+	} else {
+		if (td->filetype == FIO_TYPE_CHAR)
+			flags |= O_RDWR;
+		else
+			flags |= O_RDONLY;
+
+		td->fd = open(td->file_name, flags);
 	}
 
 	if (td->fd == -1) {
@@ -1609,23 +1772,19 @@ static void init_disk_util(struct thread_data *td)
 		return;
 
 	/*
-	 * for md/dm, there's no queue dir. we already have the right place
+	 * If there's a ../queue/ directory there, we are inside a partition.
+	 * Check if that is the case and jump back. For loop/md/dm etc we
+	 * are already in the right spot.
 	 */
-	sprintf(tmp, "%s/stat", foo);
-	if (stat(tmp, &st)) {
-		/*
-		 * if this is inside a partition dir, jump back to parent
-		 */
-		sprintf(tmp, "%s/queue", foo);
+	sprintf(tmp, "%s/../queue", foo);
+	if (!stat(tmp, &st)) {
+		p = dirname(foo);
+		sprintf(tmp, "%s/queue", p);
 		if (stat(tmp, &st)) {
-			p = dirname(foo);
-			sprintf(tmp, "%s/queue", p);
-			if (stat(tmp, &st)) {
-				fprintf(stderr, "unknown sysfs layout\n");
-				return;
-			}
-			sprintf(foo, "%s", p);
+			fprintf(stderr, "unknown sysfs layout\n");
+			return;
 		}
+		sprintf(foo, "%s", p);
 	}
 
 	disk_util_add(dev, foo);
@@ -1670,13 +1829,17 @@ static void update_rusage_stat(struct thread_data *td)
 static void *thread_main(void *data)
 {
 	struct thread_data *td = data;
-	int ret = 1;
 
 	if (!td->use_thread)
 		setsid();
 
 	td->pid = getpid();
 
+	INIT_LIST_HEAD(&td->io_u_freelist);
+	INIT_LIST_HEAD(&td->io_u_busylist);
+	INIT_LIST_HEAD(&td->io_hist_list);
+	INIT_LIST_HEAD(&td->io_log_list);
+
 	if (init_io_u(td))
 		goto err;
 
@@ -1688,6 +1851,9 @@ static void *thread_main(void *data)
 	if (init_io(td))
 		goto err;
 
+	if (init_iolog(td))
+		goto err;
+
 	if (td->ioprio) {
 		if (ioprio_set(IOPRIO_WHO_PROCESS, 0, td->ioprio) == -1) {
 			td_verror(td, errno);
@@ -1695,15 +1861,21 @@ static void *thread_main(void *data)
 		}
 	}
 
+	if (nice(td->nice) < 0) {
+		td_verror(td, errno);
+		goto err;
+	}
+
+	if (init_random_state(td))
+		goto err;
+
+	td_set_runstate(td, TD_INITIALIZED);
 	sem_post(&startup_sem);
 	sem_wait(&td->mutex);
 
 	if (!td->create_serialize && setup_file(td))
 		goto err;
 
-	if (init_random_state(td))
-		goto err;
-
 	gettimeofday(&td->epoch, NULL);
 
 	while (td->loops--) {
@@ -1720,6 +1892,9 @@ static void *thread_main(void *data)
 		do_io(td);
 
 		td->runtime[td->ddir] += mtime_since_now(&td->start);
+		if (td_rw(td) && td->io_bytes[td->ddir ^ 1])
+			td->runtime[td->ddir ^ 1] = td->runtime[td->ddir];
+
 		update_rusage_stat(td);
 
 		if (td->error || td->terminate)
@@ -1739,14 +1914,14 @@ static void *thread_main(void *data)
 			break;
 	}
 
-	ret = 0;
-
 	if (td->bw_log)
 		finish_log(td, td->bw_log, "bw");
 	if (td->slat_log)
 		finish_log(td, td->slat_log, "slat");
 	if (td->clat_log)
 		finish_log(td, td->clat_log, "clat");
+	if (td->write_iolog)
+		write_iolog_close(td);
 
 	if (exitall_on_terminate)
 		terminate_threads(td->groupid);
@@ -1760,10 +1935,6 @@ err:
 		munmap(td->mmap, td->file_size);
 	cleanup_io(td);
 	cleanup_io_u(td);
-	if (ret) {
-		sem_post(&startup_sem);
-		sem_wait(&td->mutex);
-	}
 	td_set_runstate(td, TD_EXITED);
 	return NULL;
 
@@ -1810,14 +1981,15 @@ static void show_ddir_status(struct thread_data *td, struct group_run_stats *rs,
 			     int ddir)
 {
 	char *ddir_str[] = { "read ", "write" };
-	unsigned long min, max, bw;
+	unsigned long min, max;
+	unsigned long long bw;
 	double mean, dev;
 
 	if (!td->runtime[ddir])
 		return;
 
 	bw = td->io_bytes[ddir] / td->runtime[ddir];
-	printf("  %s: io=%6lluMiB, bw=%6luKiB/s, runt=%6lumsec\n", ddir_str[ddir], td->io_bytes[ddir] >> 20, bw, td->runtime[ddir]);
+	printf("  %s: io=%6lluMiB, bw=%6lluKiB/s, runt=%6lumsec\n", ddir_str[ddir], td->io_bytes[ddir] >> 20, bw, td->runtime[ddir]);
 
 	if (calc_lat(&td->slat_stat[ddir], &min, &max, &mean, &dev))
 		printf("    slat (msec): min=%5lu, max=%5lu, avg=%5.02f, dev=%5.02f\n", min, max, mean, dev);
@@ -1844,7 +2016,8 @@ static void show_thread_status(struct thread_data *td,
 	printf("Client%d (groupid=%d): err=%2d:\n", td->thread_number, td->groupid, td->error);
 
 	show_ddir_status(td, rs, td->ddir);
-	show_ddir_status(td, rs, td->ddir ^ 1);
+	if (td->io_bytes[td->ddir ^ 1])
+		show_ddir_status(td, rs, td->ddir ^ 1);
 
 	if (td->runtime[0] + td->runtime[1]) {
 		double runt = td->runtime[0] + td->runtime[1];
@@ -1874,7 +2047,12 @@ static void check_str_update(struct thread_data *td)
 			c = 'E';
 			break;
 		case TD_RUNNING:
-			if (td_read(td)) {
+			if (td_rw(td)) {
+				if (td->sequential)
+					c = 'M';
+				else
+					c = 'm';
+			} else if (td_read(td)) {
 				if (td->sequential)
 					c = 'R';
 				else
@@ -1892,6 +2070,9 @@ static void check_str_update(struct thread_data *td)
 		case TD_CREATED:
 			c = 'C';
 			break;
+		case TD_INITIALIZED:
+			c = 'I';
+			break;
 		case TD_NOT_CREATED:
 			c = 'P';
 			break;
@@ -1937,13 +2118,23 @@ static int thread_eta(struct thread_data *td, unsigned long elapsed)
 	unsigned int eta_sec = 0;
 
 	bytes_total = td->total_io_size;
-	if (td->verify)
-		bytes_total <<= 1;
+
+	/*
+	 * if writing, bytes_total will be twice the size. If mixing,
+	 * assume a 50/50 split and thus bytes_total will be 50% larger.
+	 */
+	if (td->verify) {
+		if (td_rw(td))
+			bytes_total = bytes_total * 3 / 2;
+		else
+			bytes_total <<= 1;
+	}
 	if (td->zone_size && td->zone_skip)
 		bytes_total /= (td->zone_skip / td->zone_size);
 
 	if (td->runstate == TD_RUNNING || td->runstate == TD_VERIFYING) {
 		double perc;
+
 		bytes_done = td->io_bytes[DDIR_READ] + td->io_bytes[DDIR_WRITE];
 		perc = (double) bytes_done / (double) bytes_total;
 		if (perc > 1.0)
@@ -1951,9 +2142,10 @@ static int thread_eta(struct thread_data *td, unsigned long elapsed)
 
 		eta_sec = (elapsed * (1.0 / perc)) - elapsed;
 
-		if (eta_sec > (td->timeout - elapsed))
+		if (td->timeout && eta_sec > (td->timeout - elapsed))
 			eta_sec = td->timeout - elapsed;
-	} else if (td->runstate == TD_NOT_CREATED || td->runstate == TD_CREATED) {
+	} else if (td->runstate == TD_NOT_CREATED || td->runstate == TD_CREATED
+			|| td->runstate == TD_INITIALIZED) {
 		int t_eta = 0, r_eta = 0;
 
 		/*
@@ -2007,6 +2199,8 @@ static void print_thread_status(void)
 
 		if (elapsed >= 3)
 			eta_secs[i] = thread_eta(td, elapsed);
+		else
+			eta_secs[i] = INT_MAX;
 
 		check_str_update(td);
 	}
@@ -2026,10 +2220,7 @@ static void print_thread_status(void)
 		}
 	}
 
-	if (eta_sec == INT_MAX)
-		eta_sec = 0;
-
-	if (eta_sec) {
+	if (eta_sec != INT_MAX && elapsed) {
 		perc = (double) elapsed / (double) (elapsed + eta_sec);
 		eta_to_str(eta_str, eta_sec);
 	}
@@ -2037,7 +2228,7 @@ static void print_thread_status(void)
 	printf("Threads now running (%d)", nr_running);
 	if (m_rate || t_rate)
 		printf(", commitrate %d/%dKiB/sec", t_rate, m_rate);
-	if (eta_sec) {
+	if (eta_sec != INT_MAX) {
 		perc *= 100.0;
 		printf(": [%s] [%3.2f%% done] [eta %s]", run_str, perc,eta_str);
 	}
@@ -2075,11 +2266,60 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 	}
 }
 
+static void fio_unpin_memory(void *pinned)
+{
+	if (pinned) {
+		if (munlock(pinned, mlock_size) < 0)
+			perror("munlock");
+		munmap(pinned, mlock_size);
+	}
+}
+
+static void *fio_pin_memory(void)
+{
+	long pagesize, pages;
+	void *ptr;
+
+	if (!mlock_size)
+		return NULL;
+
+	/*
+	 * Don't allow mlock of more than real_mem-128MB
+	 */
+	pagesize = sysconf(_SC_PAGESIZE);
+	pages = sysconf(_SC_PHYS_PAGES);
+	if (pages != -1 && pagesize != -1) {
+		unsigned long long real_mem = pages * pagesize;
+
+		if ((mlock_size + 128 * 1024 * 1024) > real_mem) {
+			mlock_size = real_mem - 128 * 1024 * 1024;
+			printf("fio: limiting mlocked memory to %lluMiB\n",
+							mlock_size >> 20);
+		}
+	}
+
+	ptr = mmap(NULL, mlock_size, PROT_READ | PROT_WRITE, MAP_PRIVATE | OS_MAP_ANON, 0, 0);
+	if (!ptr) {
+		perror("malloc locked mem");
+		return NULL;
+	}
+	if (mlock(ptr, mlock_size) < 0) {
+		munmap(ptr, mlock_size);
+		perror("mlock");
+		return NULL;
+	}
+
+	return ptr;
+}
+
 static void run_threads(void)
 {
 	struct thread_data *td;
 	unsigned long spent;
 	int i, todo, nr_running, m_rate, t_rate, nr_started;
+	void *mlocked_mem;
+
+	mlocked_mem = fio_pin_memory();
 
 	printf("Starting %d thread%s\n", thread_number, thread_number > 1 ? "s" : "");
 	fflush(stdout);
@@ -2116,6 +2356,10 @@ static void run_threads(void)
 	gettimeofday(&genesis, NULL);
 
 	while (todo) {
+		struct thread_data *map[MAX_JOBS];
+		struct timeval this_start;
+		int this_jobs = 0, left;
+
 		/*
 		 * create threads (TD_NOT_CREATED -> TD_CREATED)
 		 */
@@ -2144,9 +2388,13 @@ static void run_threads(void)
 			if (td->stonewall && (nr_started || nr_running))
 				break;
 
+			/*
+			 * Set state to created. Thread will transition
+			 * to TD_INITIALIZED when it's done setting up.
+			 */
 			td_set_runstate(td, TD_CREATED);
+			map[this_jobs++] = td;
 			sem_init(&startup_sem, 0, 1);
-			todo--;
 			nr_started++;
 
 			if (td->use_thread) {
@@ -2165,12 +2413,53 @@ static void run_threads(void)
 		}
 
 		/*
-		 * start created threads (TD_CREATED -> TD_RUNNING)
+		 * Wait for the started threads to transition to
+		 * TD_INITIALIZED.
 		 */
+		printf("fio: Waiting for threads to initialize...\n");
+		gettimeofday(&this_start, NULL);
+		left = this_jobs;
+		while (left) {
+			if (mtime_since_now(&this_start) > JOB_START_TIMEOUT)
+				break;
+
+			usleep(100000);
+
+			for (i = 0; i < this_jobs; i++) {
+				td = map[i];
+				if (!td)
+					continue;
+				if (td->runstate == TD_INITIALIZED) {
+					map[i] = NULL;
+					left--;
+				} else if (td->runstate >= TD_EXITED) {
+					map[i] = NULL;
+					left--;
+					todo--;
+					nr_running++; /* work-around... */
+				}
+			}
+		}
+
+		if (left) {
+			fprintf(stderr, "fio: %d jobs failed to start\n", left);
+			for (i = 0; i < this_jobs; i++) {
+				td = map[i];
+				if (!td)
+					continue;
+				kill(td->pid, SIGTERM);
+			}
+			break;
+		}
+
+		/*
+		 * start created threads (TD_INITIALIZED -> TD_RUNNING).
+		 */
+		printf("fio: Go for launch\n");
 		for (i = 0; i < thread_number; i++) {
 			td = &threads[i];
 
-			if (td->runstate != TD_CREATED)
+			if (td->runstate != TD_INITIALIZED)
 				continue;
 
 			td_set_runstate(td, TD_RUNNING);
@@ -2178,6 +2467,7 @@ static void run_threads(void)
 			nr_started--;
 			m_rate += td->ratemin;
 			t_rate += td->rate;
+			todo--;
 			sem_post(&td->mutex);
 		}
 
@@ -2193,6 +2483,7 @@ static void run_threads(void)
 	}
 
 	update_io_ticks();
+	fio_unpin_memory(mlocked_mem);
 }
 
 static void show_group_stats(struct group_run_stats *rs, int id)
@@ -2243,7 +2534,7 @@ static void show_run_stats(void)
 	}
 
 	for (i = 0; i < thread_number; i++) {
-		unsigned long rbw, wbw;
+		unsigned long long rbw, wbw;
 
 		td = &threads[i];
 
@@ -2265,9 +2556,9 @@ static void show_run_stats(void)
 
 		rbw = wbw = 0;
 		if (td->runtime[0])
-			rbw = td->io_bytes[0] / td->runtime[0];
+			rbw = td->io_bytes[0] / (unsigned long long) td->runtime[0];
 		if (td->runtime[1])
-			wbw = td->io_bytes[1] / td->runtime[1];
+			wbw = td->io_bytes[1] / (unsigned long long) td->runtime[1];
 
 		if (rbw < rs->min_bw[0])
 			rs->min_bw[0] = rbw;
