@@ -38,6 +38,7 @@ struct io_piece {
 	struct list_head list;
 	unsigned long long offset;
 	unsigned int len;
+	int ddir;
 };
 
 /*
@@ -115,6 +116,7 @@ struct thread_data {
 	volatile int runstate;
 	volatile int old_runstate;
 	unsigned int ddir;
+	unsigned int iomix;
 	unsigned int ioprio;
 	unsigned int sequential;
 	unsigned int bs;
@@ -132,6 +134,7 @@ struct thread_data {
 	unsigned int bw_avg_time;
 	unsigned int create_serialize;
 	unsigned int create_fsync;
+	unsigned int end_fsync;
 	unsigned int loops;
 	unsigned long long file_size;
 	unsigned long long real_file_size;
@@ -146,6 +149,17 @@ struct thread_data {
 	unsigned int use_thread;
 	unsigned int iodepth;
 	os_cpu_mask_t cpumask;
+	unsigned int jobnum;
+	unsigned int iolog;
+	unsigned int read_iolog;
+	unsigned int write_iolog;
+	unsigned int rwmixcycle;
+	unsigned int rwmixread;
+	unsigned int nice;
+
+	char iolog_file[256];
+	void *iolog_buf;
+	FILE *iolog_f;
 
 	struct drand48_data bsrange_state;
 	struct drand48_data verify_state;
@@ -215,7 +229,12 @@ struct thread_data {
 	unsigned int do_disk_util;
 	unsigned int override_sync;
 
+	struct drand48_data rwmix_state;
+	struct timeval rwmix_switch;
+	int rwmix_ddir;
+
 	struct list_head io_hist_list;
+	struct list_head io_log_list;
 };
 
 #define td_verror(td, err)						\
@@ -271,10 +290,12 @@ enum {
 	FIO_LIBAIO	= 1 << 2,
 	FIO_POSIXAIO	= 1 << 3,
 	FIO_SGIO	= 1 << 4,
+	FIO_SPLICEIO	= 1 << 5 | FIO_SYNCIO,
 };
 
 #define td_read(td)		((td)->ddir == DDIR_READ)
 #define td_write(td)		((td)->ddir == DDIR_WRITE)
+#define td_rw(td)		((td)->iomix != 0)
 
 #define BLOCKS_PER_MAP		(8 * sizeof(long))
 #define TO_MAP_BLOCK(td, b)	((b) - ((td)->file_offset / (td)->min_bs))
